package org.jbpm.test;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import org.jbpm.api.Execution;
import org.jbpm.api.activity.ActivityBehaviour;
import org.jbpm.api.activity.ActivityExecution;
import org.jbpm.pvm.internal.model.Activity;
import org.jbpm.pvm.internal.model.Condition;
import org.jbpm.pvm.internal.model.ExecutionImpl;
import org.jbpm.pvm.internal.model.Transition;

public abstract class AbstractRepeatableForkActivity implements ActivityBehaviour {

    private AtomicInteger repeatNumber = new AtomicInteger(-1);

    protected enum RepeatInfo {
        NOT,
        REPEAT,
        LAST_REPEAT;
    }

    public void execute(ActivityExecution ae) throws Exception {
        this.setUp(ae);
        execute((ExecutionImpl) ae);
        this.cleanUp(ae);
    }

    public void execute(ExecutionImpl execution)  throws Exception {
        Activity activity = execution.getActivity();
        List<Transition> forkingTransitions = new ArrayList<Transition>();
        List<Transition> outgoingTransitions = activity.getOutgoingTransitions();
        for (Transition transition : outgoingTransitions) {
            Condition condition = transition.getCondition();
            if ((condition == null) || (condition.evaluate(execution))) {
                forkingTransitions.add(transition);
            }
        }

        if (forkingTransitions.size() == 0) {
            execution.end();
        } else {
            ExecutionImpl concurrentRoot = null;
            if (Execution.STATE_ACTIVE_ROOT.equals(execution.getState())) {
                concurrentRoot = execution;
                execution.setState(Execution.STATE_INACTIVE_CONCURRENT_ROOT);
                execution.setActivity(null);
            } else if (Execution.STATE_ACTIVE_CONCURRENT.equals(execution.getState())) {
                concurrentRoot = execution.getParent();
            }
           
            for (RepeatInfo repeatInfo = this.shouldRepeat();
                    repeatInfo != RepeatInfo.NOT; repeatInfo = this.shouldRepeat()) {

                Map<String, ?> variables = this.getVariablesForRepeat();

                if(this.shouldRepeat() == RepeatInfo.NOT)
                    repeatInfo = RepeatInfo.LAST_REPEAT;
                
                final int transitionLength = forkingTransitions.size();               
                for (int i = 0; i < transitionLength; i++) {
                    Transition transition = forkingTransitions.get(i);
                    String childExecutionName = transition.getName();
                    ExecutionImpl concurrentExecution =
                            concurrentRoot.createExecution(
                            childExecutionName+repeatNumber.incrementAndGet());

                    
                    if (variables != null) {
                        for (Map.Entry<String, ?> entry : variables.entrySet()) {
                            concurrentExecution.setVariable(entry.getKey(),
                                    entry.getValue());
                        }
                    }
                    concurrentExecution.setActivity(activity);
                    concurrentExecution.setState(Execution.STATE_ACTIVE_CONCURRENT);

                    if(repeatInfo == RepeatInfo.LAST_REPEAT &&
                            i == (transitionLength - 1)){

                        concurrentExecution.setVariable(
                                ActivityShareInfo.JoinCompleteIndicator.name,
                                ActivityShareInfo.JoinCompleteIndicator.value);

                        repeatInfo = RepeatInfo.NOT;
                    }
                    concurrentExecution.take(transition);

                    if (concurrentRoot.isEnded()) {
                        break;
                    }
                }
            }
        }
    }

    protected abstract void setUp(ActivityExecution ae);
    
    protected abstract void cleanUp(ActivityExecution ae);

    protected abstract RepeatInfo shouldRepeat();

    protected abstract Map<String, ?> getVariablesForRepeat();
    
}
